package aexp.samplerecognizer;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;

import android.media.AudioFormat;
import android.media.AudioRecord;
import android.media.MediaRecorder;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.os.SystemClock;
import android.app.Activity;
import android.support.v8.renderscript.Allocation;
import android.support.v8.renderscript.RenderScript;
import android.support.v8.renderscript.Element;
import android.support.v8.renderscript.Script;
import android.util.Log;
import android.view.Menu;
import android.view.View;
import android.widget.Button;
import android.widget.SeekBar;
import android.widget.TextView;

public class MainActivity extends Activity implements SeekBar.OnSeekBarChangeListener {

	enum RecordingState {
		IDLE,
		RECORD_REF,
		RECORD
	}
	static final int PARM_SIZE = 5;
	static final int DIAGLEN = 0;
	static final int DIAGX = 1;
	static final int DIAGY = 2;
	static final int DIAGNO = 3;
	static final int DPERMUTATIONINDEX = 4;
	static final int SAMPLERATE = 44100;
	static final int POWER_WINDOW_LEN = 40;
	static final String LOG_TAG = "SampleRecognizer";
	static final boolean EXTERNAL_FILE = true;
	static final boolean DEBUG_FILE = true;
	static final String FILE_REFNAME = "ref.raw";
	static final String FILE_SAMPLESNAME = "samples.raw";
	static final String FILE_DIAG_MEAS = "diagmeas.csv";
	static final int PARALLEL_LIMIT = 5000;
	static final int PARALLEL_LIMIT_SPIN_UP = 4900;

	
	@Override
	protected void onCreate(Bundle savedInstanceState) {
		recordingState = RecordingState.IDLE;
		recorder = null;
		uiHandler = new Handler();
		super.onCreate(savedInstanceState);
		setContentView(R.layout.activity_main);
		btnRecordRef = (Button) findViewById(R.id.btn_recordref);
        btnRecordRef.setOnClickListener(new View.OnClickListener() {
            public void onClick(View view) {
            	recordRef();
     		}
        });
        btnRecord = (Button) findViewById(R.id.btn_record);
        btnRecord.setOnClickListener(new View.OnClickListener() {
            public void onClick(View view) {
            	record();
     		}
        });
        tvStatus = (TextView) findViewById(R.id.tv_status);
        btnStop = (Button) findViewById(R.id.btn_stop);
        btnStop.setOnClickListener(new View.OnClickListener() {
            public void onClick(View view) {
            	stopRecord();
            }
        });
        tvDistance = (TextView) findViewById(R.id.tv_distance);
        sbDistance = (SeekBar) findViewById(R.id.sb_distance);
        sbDistance.setOnSeekBarChangeListener( this );
        updateLimitDistanceFromSeekBar();
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		// Inflate the menu; this adds items to the action bar if it is present.
		getMenuInflater().inflate(R.menu.main, menu);
		return true;
	}
	
	@Override
	public void onProgressChanged(SeekBar seekBar, int progress,
			boolean fromUser) {
		updateLimitDistanceFromSeekBar();
	}

	@Override
	public void onStartTrackingTouch(SeekBar seekBar) {
	}

	@Override
	public void onStopTrackingTouch(SeekBar seekBar) {
	}

	private void updateLimitDistanceFromSeekBar() {
		int pos = sbDistance.getProgress();
		pos += 300;
		tvDistance.setText( Integer.toString( pos ));
		limitDistance = pos;
	}
	
	private void recordRef() {
		btnRecordRef.setClickable(false);
		btnRecord.setClickable(false);
		btnStop.setVisibility(View.VISIBLE);
		tvStatus.setText(R.string.text_recordingref);
		recordingState = RecordingState.RECORD_REF;
		initializeRecording( FILE_REFNAME,false);
	}

	private void record() {
		btnRecordRef.setClickable(false);
		btnRecord.setClickable(false);
		btnStop.setVisibility(View.VISIBLE);
		tvStatus.setText(R.string.text_recording);
		recordingState = RecordingState.RECORD;
		initializeRecording( FILE_SAMPLESNAME,true);
	}
	
	private void stopRecord() {
		if( audioReaderThread != null ) {
			audioReaderThread.interrupt();
			audioReaderThread = null;
		}
	}
	
	private void stopRecordUI(String msg) {
		btnRecordRef.setClickable(true);
		btnRecord.setClickable(true);
		btnStop.setVisibility(View.INVISIBLE);
		if( msg != null )
			tvStatus.setText(msg);
		else
			tvStatus.setText( R.string.text_idle);
		recordingState = RecordingState.IDLE;		
	}
		
	private void initializeRecording( String fname, boolean needsAnalysis) {
		bufferSize = AudioRecord.getMinBufferSize(
				SAMPLERATE,
				AudioFormat.CHANNEL_IN_MONO,
				AudioFormat.ENCODING_PCM_16BIT)*4;
		Log.d( LOG_TAG, "bufferSize: "+bufferSize);
		recorder = new AudioRecord( 
				MediaRecorder.AudioSource.MIC,
				SAMPLERATE,
				AudioFormat.CHANNEL_IN_MONO,
				AudioFormat.ENCODING_PCM_16BIT,
				bufferSize
				);
		audioBuffer = new byte[ bufferSize ];
		fos = null;
		if( EXTERNAL_FILE ) {
			File f = new File( Environment.getExternalStorageDirectory(),fname );
			try {
				fos = new FileOutputStream( f );
			} catch (FileNotFoundException e) {
				String errMsg = "Cannot open file: "+f.getAbsolutePath();
				Log.e( LOG_TAG, errMsg ,e );
				stopRecordUI( errMsg );
			}
		}
		else {
			try {
				fos = openFileOutput( fname, MODE_PRIVATE);
			} catch (FileNotFoundException e) {
				String errMsg = "Cannot open file: "+fname;
				Log.e( LOG_TAG, errMsg ,e );
				stopRecordUI( errMsg );
			}
		}
		audioReaderThread = new AudioReaderThread(needsAnalysis);
		audioReaderThread.start();
	}
	
	private int findReferenceSignalJava() throws IOException {
		short refSignal[] = readSignal( FILE_REFNAME );
		double refPower = calculateMaxPower( refSignal, POWER_WINDOW_LEN );
		short recSignal[] = readSignal( FILE_SAMPLESNAME);
		scaleSignalByPower( recSignal,refPower, POWER_WINDOW_LEN );
		long t1 = System.currentTimeMillis();
		int d = distanceLM( refSignal,recSignal);
		long t2 = System.currentTimeMillis();
		dt = t2 - t1;
		Log.d( LOG_TAG,"distanceLM: timestamp: "+dt);
		Log.d( LOG_TAG, "distance (Java): "+d);
		return d;
	}

	private int findReferenceSignalJavaF() throws IOException {
		short refSignal[] = readSignal( FILE_REFNAME );
		double refPower = calculateMaxPower( refSignal, POWER_WINDOW_LEN );
		short recSignal[] = readSignal( FILE_SAMPLESNAME);
		scaleSignalByPower( recSignal,refPower, POWER_WINDOW_LEN );
		long t1 = System.currentTimeMillis();
		int s1len = refSignal.length;
		int s2len = recSignal.length;
		int d0[] = new int[s1len];
		int d1[] = new int[s1len];
		int d = distanceLMF( refSignal,recSignal,d0,d1,s1len,s2len);
		long t2 = System.currentTimeMillis();
		dt = t2 - t1;
		Log.d( LOG_TAG,"distanceLM: timestamp: "+dt);
		Log.d( LOG_TAG, "distance (JavaF): "+d);
		return d;
	}
	
	private int kernel(int s1,int s2,int d0xm1,int d1xm1,int d0x) {
		int cs = signalCost( s1,s2);
		int m = Math.min(d0xm1,d1xm1 );
		m = Math.min( m, d0x);
		return cs+m;
	}

	
	private int findReferenceSignalC99Parallel() throws IOException {
		short signal1[] = readSignal( FILE_REFNAME );
		double refPower = calculateMaxPower( signal1, POWER_WINDOW_LEN );
		short signal2[] = readSignal( FILE_SAMPLESNAME);
		scaleSignalByPower( signal2,refPower, POWER_WINDOW_LEN );
/*
		Object ts[] = testSignals();
		short signal1[] = (short[])ts[0];
		short signal2[] = (short[])ts[1];
		int refDistance = distance(signal1,signal2);
		int refDistance2 = distanceLM(signal1,signal2);
		int d0lmf[] = new int[ signal1.length];
		int d1lmf[] = new int[ signal1.length];
		int refDistance3 = distanceLMF(signal1,signal2,d0lmf,d1lmf,signal1.length,signal2.length);
*/
		long t1 = System.currentTimeMillis();
		int n = signal1.length;
		int m = signal2.length;
		if( n < m ) {
			short tmp[] = signal1;
			signal1 = signal2;
			signal2 = tmp;
			n = signal1.length;
			m = signal2.length;
		}
		int minnm = Math.min(n,m);
		int maxnm = Math.max(n,m);
		int totalDiags = n+m-1;
		RenderScript rsC = RenderScript.create( this);
		ScriptC_dtwparallel script = new ScriptC_dtwparallel( 
				rsC, 
				getResources(), 
				R.raw.dtwparallel);
		script.set_s1len( signal1.length);
		Allocation signal1Allocation = Allocation.createSized(
				rsC,
				Element.I16(rsC),
				n);
		signal1Allocation.copyFrom(signal1);
		script.bind_signal1(signal1Allocation);
		script.set_s2len( signal2.length);
		Allocation signal2Allocation = Allocation.createSized(
				rsC,
				Element.I16(rsC),
				m);
		signal2Allocation.copyFrom(signal2);
		script.bind_signal2(signal2Allocation);
		Allocation d0Allocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				minnm);
		Allocation d1Allocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				minnm);
		Allocation d2Allocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				minnm);
		Allocation parmAllocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				PARM_SIZE);
		int parmArray[] = new int[PARM_SIZE];
		Allocation d0 = d0Allocation;
		Allocation d1 = d1Allocation;
		Allocation d2 = d2Allocation;
// Upper left corner: DTW[0][0] = signalCost(signal1[0],signal2[0])
		int d0init[] = new int[minnm];
		d0init[0] = signalCost( (int)signal1[0],(int)signal2[0] );
		d0.copyFrom(d0init);
		int d1init[] = new int[minnm];
		d1init[0] = d0init[0] + signalCost( (int)signal1[1],(int)signal2[0] );
		d1init[1] = d0init[0] + signalCost( (int)signal1[0],(int)signal2[1] );
		d1.copyFrom(d1init);
		Script.LaunchOptions sc = new Script.LaunchOptions();
// There are 4 diags that consist of either 1 or 2 elements (in two opposite corners of the matrix)
// Executing the parallel kernel for these rows is pointless. That's why we start from diag #2 and end
// at totalDiags-2
		script.set_minnm(minnm);
		script.set_maxnm(maxnm);
		ArrayList<DiagMeas> dm = new ArrayList<DiagMeas>();
		long fdt1 = 0L;
		long fdt2 = 0L;
		script.bind_d0a(d0Allocation);
		script.bind_d1a(d1Allocation);
		script.bind_d2a(d2Allocation);
		script.bind_parms(parmAllocation);
		d0 = d0Allocation;
		d1 = d1Allocation;
		d2 = d2Allocation;
		int dPermutationIndex = 0;
		for( int i = 2 ; i < totalDiags - 2 ; ++i ) {
			long dt1 = 0L;
// The length of the diag is the following:
// i+1 if diag# < minnm
// minnm if diag# >= minnm and diag# < maxnm
// totalDiags-diag# if diag# > maxnm
			int diagLen = 0;
			if( i < minnm )
				diagLen = i+1;
			else
			if( i < maxnm )
				diagLen = minnm;
			else
				diagLen = totalDiags - i;
// Starting X position of the diag is
// diag# if diag# < maxnm
// maxnm-1 if diag# >= maxmn
			int diagX = i >= maxnm ? maxnm - 1 : i;
// Starting Y position of the diag is
// 0 if diag# < maxnm
// diag# - maxnm + 1 if diag# >= maxnm
			int diagY = i < maxnm ? 0 : i - maxnm + 1;
			fdt1 = SystemClock.elapsedRealtimeNanos();
			parmArray[DIAGLEN] = diagLen;
			parmArray[DIAGX] = diagX;
			parmArray[DIAGY] = diagY;
			parmArray[DIAGNO] = i;
			parmArray[DPERMUTATIONINDEX] = dPermutationIndex;
			parmAllocation.copyFrom(parmArray);
//			Log.d(LOG_TAG,"diag#"+i+"; diagLen: "+diagLen+"; diagX: "+diagX+"; diagY: "+diagY);
			sc.setX(0, diagLen);
			fdt2 = SystemClock.elapsedRealtimeNanos();
			long dt2 = fdt2 - fdt1;
			fdt1 = SystemClock.elapsedRealtimeNanos();
			script.forEach_dtw(d1, d2, sc);
			fdt2 = SystemClock.elapsedRealtimeNanos();
			long dt3 = fdt2 - fdt1;
			DiagMeas dmt = new DiagMeas();
			dmt.diagLen = diagLen;
			dmt.dt1 = dt1;
			dmt.dt2 = dt2;
			dmt.dt3 = dt3;
			dm.add( dmt );
//			d2.copyTo(d0init);
//			dumpArray( d0init,diagLen);
			dPermutationIndex = (dPermutationIndex+1) % 3;
			Allocation tmp = d0;
			d0 = d1;
			d1 = d2;
			d2 = tmp;
		}
		rsC.finish();
// Finish the two last diags
// last-1 diag: 2 elements
		int d0a[] = new int[minnm];
		d0.copyTo(d0a);
		int d1a[] = new int[minnm];
		d1.copyTo(d1a);
		int dtwnmm1 = kernel( (int)signal1[n-1],
								(int)signal2[m-2],
								d0a[1],
								d1a[0],
								d1a[1]);
		int dtwnmm2 = kernel( (int)signal1[n-2],
								(int)signal2[m-1],
								d0a[2],
								d1a[1],
								d1a[2]);
// last diag: 1 element (result)
		int maxc = kernel( (int)signal1[n-1],
							(int)signal2[m-1],
							d1a[1],
							dtwnmm1,
							dtwnmm2);
		long t2 = System.currentTimeMillis();
		dt = t2 - t1;
		Log.d( LOG_TAG,"invoke_dtwparallel: timestamp: "+dt);
		Log.d(LOG_TAG, "maxc: "+maxc);
		maxc /= maxnm;
		Log.d( LOG_TAG, "distance (C99Parallel): "+maxc);
		rsC.destroy();
		dumpDiagMeas( dm );
		return maxc;
	}

	private void processRow( 
			int diagNo,
			int startOffs,
			int len,
			short signal1[],
			short signal2[],
			int d0[],
			int d1[],
			int d2[],
			int minnm,
			int maxnm,
			int totalDiags,
			int diagLen) {
		// Starting X position of the diag is
		// diag# if diag# < maxnm
		// maxnm-1 if diag# >= maxmn
					int diagX = diagNo >= maxnm ? maxnm - 1 : diagNo;
		// Starting Y position of the diag is
		// 0 if diag# < maxnm
		// diag# - maxnm + 1 if diag# >= maxnm
					int diagY = diagNo < maxnm ? 0 : diagNo - maxnm + 1;
					for( int k = 0 ; k < diagLen ; ++k ) {
						int sum = 0;
						if( k == 0 && (  diagNo < maxnm ) ) {
							sum = d1[0] + signalCost((int)signal1[diagNo],signal2[0]);
						} else 
						if( ( k == ( diagLen - 1 ) ) && ( diagNo < minnm ) ) {
							sum = d1[diagNo-1] + signalCost((int)signal1[0],(int)signal2[diagNo]);
						} else 
						if( diagNo < maxnm ) {
							int s1 = signal1[diagX-k];
							int s2 = signal2[diagY+k];
							int cs = signalCost( s1,s2);
							sum = 
								Math.min( 
											d0[k-1],
											d1[k-1] );
							sum = Math.min( sum, d1[k] );
							sum += cs;
						} else 
						if( diagNo == maxnm ) {
							int s1 = signal1[diagX-k];
							int s2 = signal2[diagY+k];
							int cs = signalCost( s1,s2);
							sum = 
								Math.min( 
											d0[k],
											d1[k] );
							sum = Math.min( sum, d1[k+1] );
							sum += cs;
						} else 
						if( diagNo > maxnm ) {
							int s1 = signal1[diagX-k];
							int s2 = signal2[diagY+k];
							int cs = signalCost( s1,s2);
							sum = 
								Math.min( 
											d0[k+1],
											d1[k] );
							sum = Math.min( sum, d1[k+1] );
							sum += cs;
						}
						d2[k] = sum;
					}		
	}

	private native void processNativeSlow( 
			short signal1[],
			short signal2[],
			int d0[],
			int d1[],
			int d2[],
			int minnm,
			int maxnm,
			int totalDiags );
	
	private native void processNative( 
			short signal1[],
			short signal2[],
			int d0[],
			int d1[],
			int d2[],
			int minnm,
			int maxnm,
			int totalDiags );

	private native void processNativeNEON( 
			short signal1[],
			short signal2[],
			int d0[],
			int d1[],
			int d2[],
			int minnm,
			int maxnm,
			int totalDiags );
	
	
	private int findReferenceSignalJavaParallel() throws IOException {

		SampleData signal1data = readSignalPadded( FILE_REFNAME );
		double refPower = calculateMaxPower( signal1data.samples, POWER_WINDOW_LEN );
		SampleData signal2data = readSignalPadded( FILE_SAMPLESNAME);
		scaleSignalByPower( signal2data.samples,refPower, POWER_WINDOW_LEN );
/*		
		Object ts[] = testSignals();
		SampleData signal1data = new SampleData();
		short signal1[] = (short[])ts[0];
		signal1data.samples = signal1;
		signal1data.length = signal1.length;
		SampleData signal2data = new SampleData();
		short signal2[] = (short[])ts[1];
		signal2data.samples = signal2;
		signal2data.length = signal2.length;
*/
/*
		int refDistance = distance(signal1,signal2);
		int refDistance2 = distanceLM(signal1,signal2);
		int d0lmf[] = new int[ signal1.length];
		int d1lmf[] = new int[ signal1.length];
		int refDistance3 = distanceLMF(signal1,signal2,d0lmf,d1lmf,signal1.length,signal2.length);
*/
		long t1 = System.currentTimeMillis();
		int n = signal1data.length;
		int m = signal2data.length;
		if( n < m ) {
			SampleData tmp = signal1data;
			signal1data = signal2data;
			signal2data = tmp;
			n = signal1data.length;
			m = signal2data.length;
		}
		int minnm = Math.min(n,m);
		int maxnm = Math.max(n,m);
		int totalDiags = n+m-1;
		int d0[] = new int[roundSizeBy8(minnm)];
		int d1[] = new int[roundSizeBy8(minnm)];
		int d2[] = new int[roundSizeBy8(minnm)];
// Upper left corner: DTW[0][0] = signalCost(signal1[0],signal2[0])
		d0[0] = signalCost( (int)signal1data.samples[0],(int)signal2data.samples[0] );
		d1[0] = d0[0] + signalCost( (int)signal1data.samples[1],(int)signal2data.samples[0] );
		d1[1] = d0[0] + signalCost( (int)signal1data.samples[0],(int)signal2data.samples[1] );
		// There are 4 diags that consist of either 1 or 2 elements (in two opposite corners of the matrix)
		// Executing the parallel kernel for these rows is pointless. That's why we start from diag #2 and end
		// at totalDiags-2
		processNativeNEON( 
				signal1data.samples,
				signal2data.samples,
				d0,
				d1,
				d2,
				minnm,
				maxnm,
				totalDiags );

		// Finish the two last diags
		// last-1 diag: 2 elements
		int dtwnmm1 = kernel( (int)signal1data.samples[n-1],
				(int)signal2data.samples[m-2],
				d0[1],
				d1[0],
				d1[1]);
		int dtwnmm2 = kernel( (int)signal1data.samples[n-2],
				(int)signal2data.samples[m-1],
				d0[2],
				d1[1],
				d1[2]);
		// last diag: 1 element (result)
		int maxc = kernel( (int)signal1data.samples[n-1],
				(int)signal2data.samples[m-1],
				d1[1],
				dtwnmm1,
				dtwnmm2);
		long t2 = System.currentTimeMillis();
		dt = t2 - t1;
		Log.d( LOG_TAG,"invoke_dtwparallel_java: timestamp: "+dt);
		Log.d(LOG_TAG, "maxc: "+maxc);
		maxc /= maxnm;
		Log.d( LOG_TAG, "distance (JavaParallel): "+maxc);
		return maxc;
	}
		
	private int findReferenceSignalC99ParallelRSOnly() throws IOException {
		short signal1[] = readSignal( FILE_REFNAME );
		double refPower = calculateMaxPower( signal1, POWER_WINDOW_LEN );
		short signal2[] = readSignal( FILE_SAMPLESNAME);
		scaleSignalByPower( signal2,refPower, POWER_WINDOW_LEN );
/*
		Object ts[] = testSignals();
		short signal1[] = (short[])ts[0];
		short signal2[] = (short[])ts[1];
		int refDistance = distance(signal1,signal2);
		int refDistance2 = distanceLM(signal1,signal2);
		int d0lmf[] = new int[ signal1.length];
		int d1lmf[] = new int[ signal1.length];
		int refDistance3 = distanceLMF(signal1,signal2,d0lmf,d1lmf,signal1.length,signal2.length);
*/
		long t1 = System.currentTimeMillis();
		int n = signal1.length;
		int m = signal2.length;
		if( n < m ) {
			short tmp[] = signal1;
			signal1 = signal2;
			signal2 = tmp;
			n = signal1.length;
			m = signal2.length;
		}
		int minnm = Math.min(n,m);
		int maxnm = Math.max(n,m);
		RenderScript rsC = RenderScript.create( this);
		ScriptC_dtwparallel2 script = new ScriptC_dtwparallel2( 
				rsC, 
				getResources(), 
				R.raw.dtwparallel2);
		script.set_s(script);
		script.set_n( signal1.length);
		Allocation signal1Allocation = Allocation.createSized(
				rsC,
				Element.I16(rsC),
				n);
		signal1Allocation.copyFrom(signal1);
		script.bind_signal1(signal1Allocation);
		script.set_m( signal2.length);
		Allocation signal2Allocation = Allocation.createSized(
				rsC,
				Element.I16(rsC),
				m);
		signal2Allocation.copyFrom(signal2);
		script.bind_signal2(signal2Allocation);
		Allocation d0Allocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				minnm);
		Allocation d1Allocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				minnm);
		Allocation d2Allocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				minnm);
// Upper left corner: DTW[0][0] = signalCost(signal1[0],signal2[0])
		int d0init[] = new int[minnm];
		d0init[0] = signalCost( (int)signal1[0],(int)signal2[0] );
		d0Allocation.copyFrom(d0init);
		int d1init[] = new int[minnm];
		d1init[0] = d0init[0] + signalCost( (int)signal1[1],(int)signal2[0] );
		d1init[1] = d0init[0] + signalCost( (int)signal1[0],(int)signal2[1] );
		d1Allocation.copyFrom(d1init);
// There are 4 diags that consist of either 1 or 2 elements (in two opposite corners of the matrix)
// Executing the parallel kernel for these rows is pointless. That's why we start from diag #2 and end
// at totalDiags-2
		script.bind_d0a(d0Allocation);
		script.bind_d1a(d1Allocation);
		script.bind_d2a(d2Allocation);
		script.invoke_dtw();
		rsC.finish();
// Finish the two last diags
// last-1 diag: 2 elements
		Allocation d0 = null;
		Allocation d1 = null;
		int iters = n+m-1-4;
		switch( iters % 3 ) {
		case 0:
			d0 = d0Allocation;
			d1 = d1Allocation;
			break;
		case 1:
			d0 = d1Allocation;
			d1 = d2Allocation;
			break;
		case 2:
			d0 = d2Allocation;
			d1 = d0Allocation;
			break;
		}
		int d0a[] = new int[minnm];
		d0.copyTo(d0a);
		int d1a[] = new int[minnm];
		d1.copyTo(d1a);
		int dtwnmm1 = kernel( (int)signal1[n-1],
								(int)signal2[m-2],
								d0a[1],
								d1a[0],
								d1a[1]);
		int dtwnmm2 = kernel( (int)signal1[n-2],
								(int)signal2[m-1],
								d0a[2],
								d1a[1],
								d1a[2]);
// last diag: 1 element (result)
		int maxc = kernel( (int)signal1[n-1],
							(int)signal2[m-1],
							d1a[1],
							dtwnmm1,
							dtwnmm2);
		long t2 = System.currentTimeMillis();
		dt = t2 - t1;
		Log.d( LOG_TAG,"invoke_dtwparallelrsonly: timestamp: "+dt);
		Log.d(LOG_TAG, "maxc: "+maxc);
		maxc /= maxnm;
		Log.d( LOG_TAG, "distance (C99ParallelRSOnly): "+maxc);
		rsC.destroy();
		return maxc;
	}	
	
	private int distanceC99(short refSignal[],short recSignal[] ) {
		long t1 = System.currentTimeMillis();
		RenderScript rsC = RenderScript.create( this);
		ScriptC_dtw script = new ScriptC_dtw( 
				rsC, 
				getResources(), 
				R.raw.dtw);
		script.set_s1len( refSignal.length);
		Allocation signal1Allocation = Allocation.createSized(
				rsC,
				Element.I16(rsC),
				refSignal.length);
		signal1Allocation.copyFrom(refSignal);
		script.bind_signal1(signal1Allocation);
		script.set_s2len( recSignal.length);
		Allocation signal2Allocation = Allocation.createSized(
				rsC,
				Element.I16(rsC),
				recSignal.length);
		signal2Allocation.copyFrom(recSignal);
		script.bind_signal2(signal2Allocation);
		Allocation d0Allocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				refSignal.length);
		script.bind_d0(d0Allocation);
		Allocation d1Allocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				refSignal.length);
		script.bind_d1(d1Allocation);
		Allocation rAllocation = Allocation.createSized(
				rsC,
				Element.I32(rsC),
				1);
		script.bind_r(rAllocation);
		script.invoke_dtw();
		rsC.finish();
		int result[] = new int[1];
		rAllocation.copyTo(result);
		long t2 = System.currentTimeMillis();
		dt = t2 - t1;
		Log.d( LOG_TAG,"invoke_dtw: timestamp: "+dt);
		int maxc = result[0];
		Log.d(LOG_TAG, "distanceC99; maxc: "+maxc);
		int l = Math.max(recSignal.length, refSignal.length);
		maxc /= l;
		Log.d( LOG_TAG, "distance (C99): "+maxc);
		rsC.destroy();
		return maxc;		
	}

	private int findReferenceSignalC99() throws IOException {
		short refSignal[] = readSignal( FILE_REFNAME );
		double refPower = calculateMaxPower( refSignal, POWER_WINDOW_LEN );
		short recSignal[] = readSignal( FILE_SAMPLESNAME);
		scaleSignalByPower( recSignal,refPower, POWER_WINDOW_LEN );
		return distanceC99( refSignal,recSignal);
	}
	
	private Object[] testSignals() {
		short signal1[] = {1,2,3,4,5,6,7,8,9,10};
		short signal2[] = {7,6,5,4,3,2,1,2,3,4,5,6,7,8};
		Object r[] = new Object[2];
		r[0] = signal1;
		r[1] = signal2;
		return r;
	}
	
	private void dumpArray( int array[],int len ) {
		StringBuffer b = new StringBuffer();
		for( int k = 0 ; k < len ; ++k ) {
			b.append( Integer.toString( array[k]));
			b.append( " ");
		}
		Log.d( LOG_TAG, new String(b));
	}
	
	private void dumpDiagMeas( ArrayList<DiagMeas> dm) {
		try {
			File fosf = new File( Environment.getExternalStorageDirectory(),FILE_DIAG_MEAS );		
			FileWriter fos = new FileWriter( fosf );
			PrintWriter pw = new PrintWriter( fos );
			for( int i = 0 ; i < dm.size() ; ++i ) {
				DiagMeas dmt = dm.get(i);
				pw.println(dmt.diagLen+","+dmt.dt1+","+dmt.dt2+","+dmt.dt3);
			}
			pw.close();
		} catch( IOException ex ) {
			Log.e( LOG_TAG, "dumpDiagMeas",ex);
		}
	}
	
	private short[] trimSignal( 
			short signal[], 
			int scalingFactor ) {
// Search first for power maximum
		int max_power = 0;
		for( int i = 0 ; i < signal.length ; ++i ) {
			int v = (int)signal[i];
			int pw = v*v;
			if( pw > max_power )
				max_power = pw;
		}
		int start_index = 0;
		int end_index = 0;
		int limit_pw = max_power / 10;
		for( int i = 0 ; i < signal.length ; ++i ) {
			int v = (int)signal[i];
			int pw = v*v;
			if( pw > limit_pw ) {
				start_index = i;
				break;
			}
		}
		for( int i = signal.length - 1 ; i >= 0 ; --i ) {
			int v = (int)signal[i];
			int pw = v*v;
			if( pw > limit_pw ) {
				end_index = i;
				break;
			}			
		}
		int dlen = end_index-start_index + 1;
		dlen /= scalingFactor;
		short result[] = new short[dlen];
		int target = 0;
		for( int i = start_index ; i<= end_index ; i += scalingFactor ) {
			if( target < dlen )
				result[target] = signal[i];
			++target;
		}
		return result;
	}

	private void trimSignalPadded( 
			short signal[], 
			int scalingFactor,
			SampleData sd) {
// Search first for power maximum
		int max_power = 0;
		for( int i = 0 ; i < signal.length ; ++i ) {
			int v = (int)signal[i];
			int pw = v*v;
			if( pw > max_power )
				max_power = pw;
		}
		int start_index = 0;
		int end_index = 0;
		int limit_pw = max_power / 10;
		for( int i = 0 ; i < signal.length ; ++i ) {
			int v = (int)signal[i];
			int pw = v*v;
			if( pw > limit_pw ) {
				start_index = i;
				break;
			}
		}
		for( int i = signal.length - 1 ; i >= 0 ; --i ) {
			int v = (int)signal[i];
			int pw = v*v;
			if( pw > limit_pw ) {
				end_index = i;
				break;
			}			
		}
		int dlen = end_index-start_index + 1;
		dlen /= scalingFactor;
		short result[] = new short[roundSizeBy8(dlen)];
		sd.samples = result;
		sd.length = dlen;
		int target = 0;
		for( int i = start_index ; i<= end_index ; i += scalingFactor ) {
			if( target < dlen )
				result[target] = signal[i];
			++target;
		}
	}

	private double calculateMaxPower( short samples[],
			int wlen) {
		// Search first for maximum power
		PowerWindow pw = new PowerWindow( wlen );
		double maxPower = 0.0;
		for( int i = 0 ; i < samples.length ; ++i ) {
			double v = pw.putValue((double)samples[i]);
			if( v > maxPower )
				maxPower = v;
		}
		return maxPower;
	}
	
	private void scaleSignalByPower( short samples[],
			double maxPower1, 
			int wlen ) {
		double maxPower2 = calculateMaxPower( samples, wlen);
// Calculate scaling factor
		double scaling_factor = Math.sqrt( maxPower1/maxPower2);
		for( int i = 0 ; i < samples.length ; ++i ) {
			double v = scaling_factor * (double)samples[i];
			samples[i] = (short)v;
		}
	}

	
	private short[] readSignal(String fname) throws IOException {
		File fisf = null;
		if( EXTERNAL_FILE )
			fisf = new File( Environment.getExternalStorageDirectory(),fname );		
		else 
			fisf = getFileStreamPath( fname );
		FileInputStream fis = new FileInputStream( fisf );
		int flenInShorts = ((int)fisf.length()) / 2;
		short result[] = new short[flenInShorts];
		for( int i = 0 ; i < flenInShorts ; ++i ) {
			int b1 = fis.read();
			if( b1 < 0 )
				break;
			int b2 = fis.read();
			if( b2 < 0 )
				break;
			int v = b1 + ( b2 << 8 );
			if( v > 32767 )
				v = - ( 65536 - v );
			result[i] = (short)v;
		}
		fis.close();
		result = trimSignal( result,5 );
		return result;
	}

	private int roundSizeBy8( int size ) {
		int s = ( ( size / 8 ) + 1 ) * 8;
		return s;
	}

	private SampleData readSignalPadded(String fname) throws IOException {
		File fisf = null;
		SampleData sd = new SampleData();
		if( EXTERNAL_FILE )
			fisf = new File( Environment.getExternalStorageDirectory(),fname );		
		else 
			fisf = getFileStreamPath( fname );
		FileInputStream fis = new FileInputStream( fisf );
		int flenInShorts = ((int)fisf.length()) / 2;
		short result[] = new short[flenInShorts];
		for( int i = 0 ; i < flenInShorts ; ++i ) {
			int b1 = fis.read();
			if( b1 < 0 )
				break;
			int b2 = fis.read();
			if( b2 < 0 )
				break;
			int v = b1 + ( b2 << 8 );
			if( v > 32767 )
				v = - ( 65536 - v );
			result[i] = (short)v;
		}
		fis.close();		
		trimSignalPadded( result,5,sd );
		return sd;
	}

	
	int distance( short signal1[], short signal2[]) {
		int s1len = signal1.length;
		int s2len = signal2.length;
		int d[][] = new int[s1len][s2len];
		int s = 0;
		for( int i = 0 ; i < s1len ; ++i ) {
			s += signalCost((int)signal1[i],(int)signal2[0]);
			d[i][0] = s;
		}
		s = 0;
		for( int i = 0 ; i < s2len ; ++i ) {
			s += signalCost((int)signal1[0],(int)signal2[i]);
			d[0][i] = s;
		}
		for( int x = 1 ; x < s1len ; ++x )
			for( int y = 1 ; y < s2len ; ++y ) {
				s = signalCost( (int)signal1[x],(int)signal2[y]);
				int m = Math.min(d[x-1][y-1],d[x-1][y] );
				m = Math.min( m, d[x][y-1]);
				d[x][y] = s+m;
			}
		for( int x = 0 ; x < s1len ; ++x ) {
			dumpArray( d[x],s2len);
		}
		int maxc = d[s1len-1][s2len-1];
		int l = Math.max(s1len, s2len);
		maxc /= l;
		return maxc;
	}

	int distanceLM( short signal1[], short signal2[]) {
		int s1len = signal1.length;
		int s2len = signal2.length;
		int d0[] = new int[s1len];
		int d1[] = new int[s1len];
		int s = 0;
		for( int x = 0 ; x < s1len ; ++x ) {
			s += signalCost((int)signal1[x],(int)signal2[0]);
			d0[x] = s;
		}
//		dumpArray( d0,s1len);
		s = d0[0];
		for( int y = 1 ; y < s2len ; ++y ) {
			s += signalCost((int)signal1[0],(int)signal2[y]);
			d1[0] = s;
			for( int x = 1 ; x < s1len ; ++x ) {
				int cs = signalCost( (int)signal1[x],(int)signal2[y]);
				int m = Math.min(d0[x-1],d1[x-1] );
				m = Math.min( m, d0[x]);
				d1[x] = cs+m;
			}
			for( int x = 0 ; x < s1len ; ++x )
				d0[x] = d1[x];
//			dumpArray( d0,s1len);
//			System.arraycopy(d1, 0, d0, 0, s1len);
		}
		int maxc = d1[s1len-1];
		Log.d(LOG_TAG, "distanceLM: maxc: "+maxc);
		int l = Math.max(s1len, s2len);
		maxc /= l;
		return maxc;
	}
	
	
	int signalCost( int x1, int x2 ) {
		return Math.abs( x1-x2 );
	}

	int distanceLMF(final short signal1[], final short signal2[], final int d0[], final int d1[], final int s1len, final int s2len)
	{
	  int s = 0;
	  int signal2Y = signal2[0];
	 
	  for (int x = 0; x < s1len; x++)
	  {
	     d0[x] = (s += Math.abs(signal1[x] - signal2Y));
	  }
	 
	  s = d0[0];
	  for (int y = 1; y < s2len; y++)
	  {
	     signal2Y = signal2[y];
	 
	     d1[0] = ( s+= Math.abs(signal1[0] - signal2Y));
	 
	     for (int x = 1; x < s1len; x++)
	     {
	        //      cs
	        d1[x] = Math.abs(signal1[x] - signal2Y) + Math.min( Math.min(d0[x - 1], d1[x - 1]), d0[x] );
	     }
	 
	     System.arraycopy(d1, 0, d0, 0, s1len);
	  }
	  Log.d( LOG_TAG, "distanceLMF: "+d1[s1len - 1]);
	  // max c        length
	  return d1[s1len - 1] / Math.max(s1len, s2len);
	}
	
	TextView tvStatus;
	TextView tvDistance;
	SeekBar sbDistance;
	Button btnStop;
	Button btnRecordRef;
	Button btnRecord;
	RecordingState recordingState;
	AudioRecord recorder;
	int bufferSize;
	byte audioBuffer[];
	FileOutputStream fos;
	AudioReaderThread audioReaderThread;
	Handler uiHandler;
	int limitDistance;
	private long dt = 0L;
	
	class AudioReaderThread extends Thread {
		public AudioReaderThread( boolean needsAnalysis) {
			this.needsAnalysis = needsAnalysis;
		}
		
		public void run() {
			String errMsg = null;
			recorder.startRecording();
			while(!isInterrupted()) {
				int len = recorder.read(audioBuffer,0,bufferSize);
				if( len > 0 ) {
					try {
						fos.write(audioBuffer,0,len);
					} catch (IOException e) {
						Log.e(LOG_TAG, "buffer write error",e);
						errMsg = "Write error";
					}
				} else {
					errMsg = "Stopped";
					break;
				}
			}
			recorder.stop();
			recorder.release();
			try {
				fos.close();
			} catch (IOException e) {
				Log.e( LOG_TAG, e.getMessage(),e);
				errMsg = "File close error";
			}
			recorder = null;
			if( errMsg != null ) {
				StopRecordingUI statusTask = new StopRecordingUI( errMsg );
				uiHandler.post( statusTask );
			} else {
				if( needsAnalysis ) {
					UpdateStatus statusTask = new UpdateStatus( getString( R.string.find_reference),true);
					uiHandler.post( statusTask );
					String result = null;
					try {
						int distanceJavaF = 0;
						long dtJavaF = 0L;
						int distanceC99 = 0;
						long dtC99 = 0L;
						int distanceC99Parallel = 0;
						long dtC99Parallel = 0L;
						int distanceC99ParallelRSOnly = 0;
						long dtC99ParallelRSOnly = 0L;
						int distanceJavaParallel = 0;
						long dtJavaParallel = 0L;
						/*
						distanceJavaF = findReferenceSignalJavaF();
						dtJavaF = dt;
						Log.d( LOG_TAG, "Java implementation, fast, execution time: "+dtJavaF+" msec");
*/
						distanceC99 = findReferenceSignalC99();
						dtC99 = dt;
						Log.d( LOG_TAG, "C99 implementation, execution time: "+dtC99+" msec");
/*
						distanceC99Parallel = findReferenceSignalC99Parallel();
						dtC99Parallel = dt;
						Log.d( LOG_TAG, "C99 implementation, execution time: "+dtC99+" msec");
*/
/*
						distanceC99ParallelRSOnly = findReferenceSignalC99ParallelRSOnly();
						dtC99ParallelRSOnly = dt;
						Log.d( LOG_TAG, "C99 implementation (RS only), execution time: "+dtC99ParallelRSOnly+" msec");
*/
						distanceJavaParallel = findReferenceSignalJavaParallel();
						dtJavaParallel = dt;
						Log.d( LOG_TAG, "Java implementation (parallel), execution time: "+dtJavaParallel+" msec");
						boolean found = distanceC99 < limitDistance;
						result = found ?
							getString( R.string.ref_signal_found)
							:
							getString( R.string.ref_signal_not_found);
						result += "; distanceF="+Integer.toString(distanceJavaF);
						result += "; dtJavaF="+Long.toString(dtJavaF);
						result += "; distanceC99="+Integer.toString( distanceC99);
						result += "; dtC99="+Long.toString(dtC99);
						result += "; distanceC99Parallel="+Integer.toString( distanceC99Parallel);
						result += "; dtC99Parallel="+Long.toString(dtC99Parallel);
						result += "; distanceC99ParallelRSOnly="+Integer.toString( distanceC99ParallelRSOnly);
						result += "; dtC99ParallelRSOnly="+Long.toString(dtC99ParallelRSOnly);
						result += "; distanceJavaParallel="+Integer.toString( distanceJavaParallel);
						result += "; dtJavaParallel="+Long.toString(dtJavaParallel);
					} catch( IOException ex ) {
						result = ex.getMessage();
						Log.e( LOG_TAG, ex.getMessage(),ex);
					}
					StopRecordingUI stopRecordingTask = new StopRecordingUI( result );
					uiHandler.post( stopRecordingTask );					
				} else {
					StopRecordingUI statusTask = new StopRecordingUI( null );
					uiHandler.post( statusTask );					
				}
			}
		}
		
		boolean needsAnalysis;
	}
	
	class UpdateStatus implements Runnable {
		public UpdateStatus( String statusMsg, boolean removeStopButton ) {
			this.statusMsg = statusMsg;
			this.removeStopButton = removeStopButton;
		}

		@Override
		public void run() {
			if( removeStopButton )
				btnStop.setVisibility(View.INVISIBLE);
			tvStatus.setText( statusMsg);
		}
		
		String statusMsg;
		boolean removeStopButton;
	}

	class StopRecordingUI implements Runnable {
		public StopRecordingUI( String statusMsg ) {
			this.statusMsg = statusMsg;
		}
		
		public void run() {
			stopRecordUI(statusMsg);			
		}
		String statusMsg;
	}
	
	class PowerWindow {
		public PowerWindow( int wlen ) {
			this.wlen = wlen;
			window = new double[wlen];
			wptr = 0;
		}

		public double putValue( double v ) {
			window[wptr] = v*v;
			int p = wptr;
			double pw = 0.0;
			for( int i = 0 ; i < wlen ; ++i ) {
				pw += window[p];
				--p;
				if( p < 0 )
					p = wlen - 1;
			}
			pw = pw / (double)wlen;
			++wptr;
			wptr = wptr % wlen;
			return pw;
		}
		
		int wlen;
		double window[];
		int wptr;
	}

	class DiagMeas {
		public int diagLen;
		public long dt1 = -1L;
		public long dt2 = -1L;
		public long dt3 = -1L;
	}
	
	class SampleData {
		short samples[];
		int length;
	}
	
	static {
		System.loadLibrary( "cpucore");
	}
}
